import pandas as pd
import numpy as np
import os, posixpath
from PIL import Image


def get_filename(idx, base_path="files", depth=3, extension=".png"):
    """Generate filename from file index.

    Args:
        idx (_type_):the index
        base_path (str, optional): path to location where files are stored
        depth (int, optional): how many directories deep to go, can store 100**depth files
        extension (str, optional): file extension to add. Defaults to ".png".

    Returns:
        str: the filename
    """
    name = (f"%0{depth*2}d") % idx
    name = "/".join(["".join(x) for x in zip(*(iter(name),)*2)]) + extension
    if base_path:
        name = base_path + '/' + name
    return name


def flatten_table(row_data, name="data", parent="", parent_idx=0, lens={}, files_len=0, base_path="files", extension=".png", file_depth=3):
    """
    Takes row data that looks like a list of dicts with common schema and flattens out elements
    that look like lists into their own tables. Also pulls out files into their own dict with
    filenames that reference back to elements

    Args:
        row_data (list): the rows
        name (str, optional): name of the table being generated
        parent (str, optional): parent to this table
        parent_idx (int, optional): index of the parent to reference in the child
        lens (dict, optional): list of the lengths of tables 

    Returns:
        tuple of dicts: tables (names to rows) and files dict (paths to file binary data)
    """
    tables = {name: []}
    files = {}
    for row_idx, row in enumerate(row_data):
        tables[name].append({})
        if parent:  # Save index of our parent
            tables[name][-1][parent+'_idx'] = parent_idx
        for k, v in row.items():
            if isinstance(v, list) and len(v) and isinstance(v[0], dict):
                my_lens = {k: len(tables.get(k, []))+lens.get(k, 0) for k in set(lens.keys()).union(set(tables.keys()))}
                child_tables, child_files = flatten_table(v, name=k, parent=name, parent_idx=my_lens[name]-1, lens=my_lens, 
                                                          files_len=len(files)+files_len, base_path=base_path, extension=extension, 
                                                          file_depth=file_depth)
                files.update(child_files)
                for k, v in child_tables.items():
                    tables[k] = tables.get(k, []) + v
            elif isinstance(v, bytes):
                tables[name][-1][k] = get_filename(len(files)+files_len, base_path=base_path, extension=extension, depth=file_depth)
                files[tables[name][-1][k]] = v
            elif isinstance(v, np.ndarray) and len(v.shape) == 2:
                tables[name][-1][k] = get_filename(len(files)+files_len, base_path=base_path, extension=extension, depth=file_depth)
                files[tables[name][-1][k]] = v
            else:
                tables[name][-1][k] = v
    return tables, files


def save_homogenous_data(data, path):
    """
    Saves data that looks like a list of dicts with potentially lists and images in elements to
    flattened tables at the specified path

    Args:
        data (list of dicts): the data to save
        path (str): path to save it to (will make directory)
    """
    tables, files = flatten_table(data)
    os.makedirs(path, exist_ok=True)
    for k, v in tables.items():
        pd.DataFrame(v).to_csv(os.path.join(path, k+'.csv'))
    for k, v in files.items():
        fpath = os.path.join(path, k)
        os.makedirs(os.path.dirname(fpath), exist_ok=True)
        if isinstance(v, bytes):
            with open(fpath, 'wb') as f:
                f.write(v)   
        elif isinstance(v, np.ndarray) and len(v.shape) == 2:
            im = Image.fromarray(v)
            im.save(fpath, format='png')
        else:
            raise ValueError(f'Unrecognized type for saving files "{type(v)}"')
            

def load_data_and_images(path):
    """
    Loads all tables and image files from the data stored at path

    Args:
        path (str): path to the directory of data

    Returns:
        tuple: the dict of tables and dict of loaded images
    """
    tables = {}
    files = {}
    for p in os.listdir(path):
        if os.path.splitext(p)[1] == '.csv' and os.path.isfile(os.path.join(path, p)):
            tables[os.path.splitext(p)[0]] = pd.read_csv(os.path.join(path, p))
        else:
            for root, _, fs in os.walk(os.path.join(path, p)):
                for f in fs:
                    if f[0] != '.':
                        files[os.path.relpath(os.path.join(root, f), path).replace(os.sep,posixpath.sep)] = np.asarray(Image.open(os.path.join(root, f)))
    return tables, files


def pixel_coords_from_bbox(upper_left, lower_left, lower_right, image_shape):
    """
    Given bounding box of 2D points, generate grid of points corresponding to pixel coordinates
    for an image of shape image_shape

    Args:
        upper_left (1D np.array 2 elements): coordinates of upper left of image
        lower_left (1D np.array 2 elements): coordinates of lower left of image
        lower_right (1D np.array 2 elements): coordinates of lower right of image
        image_shape (tuple, 2 elements): shape of the image
        
    Returns:
        tuple of np.array: x and y coordinate arrays
    """
    x, y = np.mgrid[0:1:image_shape[0]*1j, 0:1:image_shape[1]*1j]
    mat = np.vstack((
        upper_left - lower_left,
        lower_right - lower_left
    )).T
    x, y = np.reshape(mat@np.vstack((x.ravel(), y.ravel()))+lower_left[:, None], (2, *x.shape))
    return x, y


def pixel_coords_from_row(row, files, prefix="", coords=['r', 'vr']):
    """
    Generates x and y coordinate arrays corresponding to the image stored in the pandas style row (or dict).
    The schema should have:
     - lower_left_x: x coordinate of lower left point of bounding box
     - lower_left_y: y coordinate of lower left point of bounding box
     - lower_right_x: x coordinate of lower right point of bounding box
     - lower_right_y: y coordinate of lower right point of bounding box
     - upper_left_x: x coordinate of upper left point of bounding box
     - upper_left_y: y coordinate of upper left point of bounding box
     - rho: the path to the image
    Args:
        row (dict-like): the dict / pandas row containing bounding box data
        files (dict): the dict of loaded files, to access the image
        prefix (str, optional): a label on the beginning of the density keys in the row. Defaults to "".
        coords (list of strs): a list of two coordinate names for the density image. Defaults to ['r', 'vr'].
    """
    bbox = {}
    for point in ['lower_left', 'lower_right', 'upper_left']:
        bbox[point] = np.array([row[prefix+point+'_'+coord] for coord in coords])
    bbox['upper_right'] = bbox['upper_left'] + bbox['lower_right'] - bbox['lower_left']
    return (x.T for x in pixel_coords_from_bbox(bbox['upper_left'], bbox['lower_left'], bbox['lower_right'], files[row[prefix+'rho']].T.shape))


def get_density_series_bounding_box(df, prefix="", coords=['r', 'vr']):
    """
    Get the axis-aligned bounding box from density images stored as a series of rows in the pandas dataframe

    Args:
        df (pandas dataframe): dataframe with rows that store density image information
        prefix (str, optional): a label on the beginning of the density keys in the row. Defaults to "".
        coords (list of strs): a list of two coordinate names for the density image. Defaults to ['r', 'vr'].

    Returns:
        tuple of dicts: min and max of the coordinates
    """
    df2 = df.copy()
    for coord in coords:
        df2[f'{prefix}upper_right_{coord}'] = df2[f'{prefix}upper_left_{coord}'] + df2[f'{prefix}lower_right_{coord}'] - df2[f'{prefix}lower_left_{coord}']

    mins = {coord: min(df2[f'{prefix}{pt}_{coord}'].min() for pt in ['lower_left', 'lower_right', 'upper_left', 'upper_right']) for coord in coords}
    maxes = {coord: max(df2[f'{prefix}{pt}_{coord}'].max() for pt in ['lower_left', 'lower_right', 'upper_left', 'upper_right']) for coord in coords}
    return mins, maxes



import matplotlib.pyplot as plt
import matplotlib.animation
from scipy import interpolate
import numpy as np

from .file_interfaces import get_density_series_bounding_box, pixel_coords_from_row


def make_density_animation(tables, files, output="", idx=0, bound_scale=1, screen_table='screens', param_table='data', timelike_key='spos', coords=['r', 'vr'], dpi=300, titlesize=16):
    """
    Generates an animation from matplotlib of density in the tables / files along the timelike coordinate.

    Args:
        tables (dict): the loaded tables
        files (dict): the loaded files
        output (str, optional): A location to save the animation as a file. Defaults to "".
        idx (int, optional): The index of the point the screen outputs are associated with. Defaults to 0.
        bound_scale (float, optional): Scaling factor for bounding box. Defaults to 1.
        screen_table (str, optional): name of the table containing the screen objects with densities. Defaults to 'screens'.
        param_table (str, optional): name of the table containing parameters associated with the screens. Defaults to 'data'.
        timelike_key (str, optional): name of the timelike coordinate in the screen table. Defaults to 'spos'.
        coords (list, optional): list of the coordinate names. Defaults to ['r', 'vr'].
        dpi (int, optional): dpi of the saved animation. Defaults to 300.
        titlesize (float, optional): font size of the title. Defaults to 16.

    Returns:
        animation object: the matplotlib animation
    """
    df = tables[screen_table][tables[screen_table][f'{param_table}_idx'] == idx].sort_values(timelike_key)

    # Setup the plot
    fig, ax = plt.subplots(facecolor='w', layout='constrained')
    p = [ax.contourf(np.zeros((2,2)), np.zeros((2,2)), np.zeros((2,2)), antialiased=True)]
    p[0].set_edgecolor("face")
    plt.gca().set_aspect('equal')
    
    # Make labels
    labels = {
        'r': 'Radius (mm)',
        'vr':'Radial Velocity (mrad)',
        'x': 'x (mm)',
        'y': 'y (mm)',
        'vx': 'x Velocity (mrad)',
        'vy': 'y Velocity (mrad)',
        'z': 'z (mm)'
    }
    plt.xlabel(labels.get(coords[0], ""))
    plt.ylabel(labels.get(coords[1], ""))
    
    # Find the bounding box for the images
    mins, maxes = get_density_series_bounding_box(df)
    bounds = {k: bound_scale*max((abs(mins[k]), abs(maxes[k]))) for k in mins}

    # The callback function for matplotlib
    def update(idx):
        p[0].remove()
        row = df.iloc[idx]
        x, y = pixel_coords_from_row(row, files)
        xinterp, yinterp = np.mgrid[-bounds['r']:bounds['r']:256j, -bounds['vr']:bounds['vr']:256j]
        ypred = interpolate.griddata(np.vstack((x.ravel(), y.ravel())).T, files[row['rho']].ravel(), np.vstack((xinterp.ravel(), yinterp.ravel())).T, method='linear', fill_value=0.0)
        ypred = np.reshape(ypred, xinterp.shape)
        p[0] = ax.contourf(1e3*xinterp, 1e3*yinterp, ypred)
        ax.set_title('s = %.2f m' % row['spos'], fontdict={'size': titlesize})
        return p

    ani = matplotlib.animation.FuncAnimation(fig, update, frames=len(df), interval=1/7*1000, blit=True, repeat=True)
    if output:
        ani.save(output, dpi=dpi)
    return ani



from pymoo.core.problem import Problem as PymooProblem
from paretobench import Problem
import numpy as np


class PyMooWrapper(PymooProblem):
    def __init__(self, problem: Problem):
        """
        Initialize the PyMooWrapper with a given problem object.

        Parameters
        ----------
        problem : Problem
            A problem object that follows the Problem class interface.
        """
        self.prob = problem
        
        super().__init__(
            n_var=self.prob.n_vars,
            n_obj=self.prob.n_objs,
            n_ieq_constr=self.prob.n_constraints,
            xl=self.prob.var_lower_bounds,
            xu=self.prob.var_upper_bounds,
            vtype=float
        )

    def _evaluate(self, X, out, *args, **kwargs):
        """
        Evaluate the problem for pymoo using vectorization.

        Parameters
        ----------
        X : np.ndarray
            2D array of decision variables, where each row is a solution
        out : dict
            Dictionary to store the output

        Returns
        -------
        None
        """
        # Evaluate the problem
        pop = self.prob(X)

        # Set the objectives
        out["F"] = pop.f

        # Set the constraints (if any) Note: pymoo uses the opposite definition of feasible as paretobench
        if self.prob.n_constraints > 0:
            out["G"] = -pop.g



import electronsandstuff.file_interfaces as fi
import numpy as np
import tempfile, os


def ccw(A,B,C):
    return (C[1]-A[1]) * (B[0]-A[0]) > (B[1]-A[1]) * (C[0]-A[0])


# Return true if line segments AB and CD intersect
def intersect(A, B, C, D):
    return ccw(A, C, D) != ccw(B, C, D) and ccw(A, B, C) != ccw(A, B, D)


def test_pixel_coords():
    np.random.seed(0)
    for _ in range(8):
        # Generate a random bounding box
        upper_left = np.array((-np.random.random(), np.random.random()))
        lower_left = np.array((-np.random.random(), -np.random.random()))
        lower_right = np.array((np.random.random(), -np.random.random()))
        upper_right = lower_right + (upper_left - lower_left)
        edges = (
            (lower_left, lower_right), (lower_right, upper_right), 
            (upper_right, upper_left), (upper_left, lower_left)
        )
        
        # Generate the grid of points
        X, Y = fi.pixel_coords_from_bbox(upper_left, lower_left, lower_right, (32, 32))
        
        # Check the boundaries
        np.testing.assert_allclose((X[0, 0], Y[0, 0]), lower_left)
        np.testing.assert_allclose((X[0, -1], Y[0, -1]), lower_right)
        np.testing.assert_allclose((X[-1, -1], Y[-1, -1]), upper_right)
        np.testing.assert_allclose((X[-1, 0], Y[-1, 0]), upper_left)

        # Shrink by a small amount to avoid (literal) edge cases
        scale = 1-1e-8
        X, Y = scale*X + (1-scale)*np.mean(X), scale*Y + (1-scale)*np.mean(Y)
        
        # Check all points lie inside the parallelogram
        for x, y in zip(X.ravel(), Y.ravel()):
            intersections = sum(intersect(*edge, np.array((x, y)), np.array((2, 2))) for edge in edges)
            assert intersections%2 == 1


def generate_test_subrow():
    row = {k: np.random.random() for k in "nicr"}
    row['image'] = np.random.randint(0, 2**16-1, size=(64, 64)).astype(np.uint16)
    return row


def generate_test_row():
    row = {k: np.random.random() for k in "jwop"}
    row['children'] = [generate_test_subrow() for _ in range(np.random.randint(4, 8))]
    row['image'] = np.random.randint(0, 2**16-1, size=(64, 64)).astype(np.uint16)
    return row


def test_save_load_homogenous_data():
    # Create some fake data
    np.random.seed(0)
    test_data = [generate_test_row() for _ in range(np.random.randint(32, 128))]

    with tempfile.TemporaryDirectory() as dir:
        # Save it to disk
        path = os.path.join(dir, 'test')
        fi.save_homogenous_data(test_data, path)
        
        # Load it back
        data, images = fi.load_data_and_images(path)
        
        # Run some basic checks
        assert data.keys() == {'data', 'children'}  # All tables are here
        assert all(x in images for x in data['data']['image'])  # All images are loaded back
        assert all(x in images for x in data['children']['image'])  # All images are loaded back

        for test, real in zip(data['data'].to_dict('records'), test_data):
            # Check this row
            assert all(np.isclose(test[k], real[k]) for k in real.keys() if len(k) == 1)  # Check the basic keys
            np.testing.assert_allclose(real['image'], images[test['image']])  # Test the image
            
            # Deal with the children
            child_df = data['children'][data['children']['data_idx'] == test['Unnamed: 0']]
            assert len(real['children']) == len(child_df)  # Child count
            for test_child, real_child in zip(child_df.to_dict('records'), real['children']):
                assert all(np.isclose(test_child[k], real_child[k]) for k in real_child.keys() if len(k) == 1)  # Check the basic keys
                np.testing.assert_allclose(real_child['image'], images[test_child['image']])  # Test the image



from xopt import VOCS
import numpy as np
from paretobench import Problem


class XoptProblemWrapper:
    def __init__(self, problem: Problem):
        """
        This class wraps a ParetoBench problem for use with xopt. After creation of the wrapper object from
        the problem, the Xopt VOCS object can be accessed through a class property. The wrapper object is
        also a callable and may be directly passed to the Xopt evaluator object.

        Example
        -------
        > import paretobench as pb
        > from xopt import Xopt, Evaluator
        > from xopt.generators.ga.cnsga import CNSGAGenerator
        >
        > prob = XoptProblemWrapper(pb.Problem.from_line_fmt('WFG1'))
        > population_size = 50
        > ev = Evaluator(function=prob, vectorized=True, max_workers=population_size)
        > X = Xopt(
        >        generator=CNSGAGenerator(vocs=prob.vocs, population_size=population_size),
        >        evaluator=ev,
        >        vocs=prob.vocs,
        >    )


        Parameters
        ----------
        problem : Problem
            A problem object that follows the Problem class interface.
        """
        self.prob = problem

    @property
    def vocs(self) -> VOCS:
        """Return the VOCS object."""
        # Construct the decision variables
        lbs = self.prob.var_lower_bounds
        ubs = self.prob.var_upper_bounds
        vars = {f"x{i}": [lb, ub] for i, (lb, ub) in enumerate(zip(lbs, ubs))}

        # Construct the objectives
        objs = {f"f{i}": "MINIMIZE" for i in range(self.prob.n_objs)}

        # The constraints
        constraints = {
            f"g{i}": ["GREATER_THAN", 0] for i in range(self.prob.n_constraints)
        }

        # Construct VOCS object
        return VOCS(variables=vars, objectives=objs, constraints=constraints)

    def __call__(self, input_dict: dict) -> dict:
        """
        Evaluate the problem using the dict -> dict convention for xopt.

        Parameters
        ----------
        input_dict : dict
            A dictionary containing the decision variables

        Returns
        -------
        dict
            A dictionary with the objectives and constraints
        """
        # Convert the input dictionary to a NumPy array of decision variables
        x = np.array([input_dict[f"x{i}"] for i in range(self.prob.n_vars)]).T

        # Evaluate the problem
        pop = self.prob(x)  # Pass single batch

        # Convert the result to the format expected by Xopt
        ret = {}
        ret.update({f"f{i}": pop.f[:, i] for i in range(self.prob.n_objs)})
        ret.update({f"g{i}": pop.g[:, i] for i in range(self.prob.n_constraints)})
        return ret

    def __repr__(self):
        return f"XoptProblemWrapper({self.prob.to_line_fmt()})"
